<?php

namespace OCA\GrueneGroupfolderManagement;

use GuzzleHttp\Client;
use GuzzleHttp\Exception\RequestException;
use OCA\GrueneGroupfolderManagement\Cache\CacheAwareInterface;
use OCA\GrueneGroupfolderManagement\Cache\CacheAwareTrait;
use OCA\GrueneGroupfolderManagement\Exceptions\ApiClientException;
use OCA\GrueneGroupfolderManagement\Model\Api\Group;
use OCA\GrueneGroupfolderManagement\Model\Organization;

class ApiClient implements CacheAwareInterface {
	use CacheAwareTrait;

	public const CACHE_TTL = 3600 * 24;

	/**
	 * @var string
	 */
	protected $server = 'https://api.netzbegruenung.de/';

	protected $username;

	protected $password;

	public function __construct($server = null, $username = null, $password = null) {
		$this->server = rtrim($server, '/');
		$this->username = $username;
		$this->password = $password;
	}

	/**
	 * @return array|null
	 */
	protected function json($url, array $attr = array()) {
		$hash = $url . '_' . md5(json_encode($attr));

		if ($this->cache !== null) {
			$cached_value = $this->cache->get($hash);
			if ($cached_value) {
				return json_decode($this->cache->get($hash), true);
			}
		}

		$config = [];
		if (!empty($this->username) && !empty($this->password)) {
			$config['auth'] = [$this->username, $this->password];
		}
		$client = new Client($config);

		try {
			$result = $client->get($this->server . $url, [
				'query' => $attr
			]);

			$content = $result->getBody()->getContents();

			if ($this->cache !== null) {
				$this->cache->set($hash, $content, self::CACHE_TTL);
			}

			$data = \json_decode($content, true);

			if (\json_last_error() == JSON_ERROR_NONE) {
				return $data;
			}
		} catch (RequestException $exception) {
			throw new ApiClientException($exception->getMessage(), $exception->getCode(), $exception);
		}
	}

	/**
	 * @return array|null
	 */
	public function partyRegionalchapter($key) {
		return $this->json('/v1/party/regionalchapters/' . urlencode($key));
	}

	/**
	 * @param string $search
	 * @param string $idPrefix
	 * @param string $type
	 *
	 * @return array|null
	 */
	public function partyRegionalchapters($search = null, $idPrefix = null, $type = null) {
		$query = [];

		if ($search !== null) {
			$query['filter[query]'] = $search;
		}

		if ($idPrefix !== null) {
			$query['filter[id_prefix]'] = $idPrefix;
		}

		if ($type !== null) {
			$query['filter[type]'] = $type;
		}

		return $this->json('/v1/party/regionalchapters', $query);
	}

	/**
	 * @return array|null
	 */
	public function partyOrganizations() {
		return $this->json('/v1/party/organizations');
	}

	protected function convertGroup(array $item) : Group {
		$model = new Group();
		if (isset($item['name'])) {
			$model->setName($item['name']);
		}
		if (!isset($item['slug'])) {
			$item['slug'] = md5($item['level'] . json_encode($item['external_refs'], true));
		}
		$model->setSlug($item['slug']);
		$model->setType($item['type']);
		$model->setActive($item['active']);
		$model->setLevel($item['level']);
		$model->setExternalRefs($item['external_refs']);

		return $model;
	}

	/**
	 * @return Group[]
	 */
	public function groupsByLevel(Organization $organization, $level) : iterable {
		$filter = array(
			'filter[organization_id]' => $organization->getId(),
			'filter[level]' => $level
		);

		$response = $this->json('/v1/party/groups', $filter);

		if (!isset($response['items'])) {
			yield from [];
		}

		foreach ($response['items'] as $item) {
			yield $this->convertGroup($item);
		}

		yield from [];
	}

	/**
	 * @return Group
	 */
	public function groupBySlugAndLevel(Organization $organization, $slug, $level = null) : ?Group {
		$filter = array(
			'filter[organization_id]' => $organization->getId(),
			'filter[slug]' => $slug
		);

		if ($level !== null) {
			$filter['filter[level]'] = $level;
		}

		$response = $this->json('/v1/party/groups', $filter);

		if (!isset($response['items'])) {
			return null;
		}

		foreach ($response['items'] as $item) {
			return $this->convertGroup($item);
		}

		return null;
	}

	/**
	 * @return Group[]
	 */
	public function groupsByTypeAndLevel(Organization $organization, $type, $level = null) : iterable {
		$filter = array(
			'filter[organization_id]' => $organization->getId(),
			'filter[type]' => $type
		);

		if ($level !== null) {
			$filter['filter[level]'] = $level;
		}

		$response = $this->json('/v1/party/groups', $filter);

		if (!isset($response['items'])) {
			yield from [];
		}

		foreach ($response['items'] as $item) {
			yield $this->convertGroup($item);
		}

		yield from [];
	}

	/**
	 * @return Group[]
	 */
	public function groupsBySherpaIdAndLevel(Organization $organization, $key, $level = null) : iterable {
		$filter = array(
			'filter[organization_id]' => $organization->getId(),
			'filter[external_refs.type]' => 'SHERPA',
			'filter[external_refs.key]' => $key
		);

		if ($level !== null) {
			$filter['filter[level]'] = $level;
		}

		$response = $this->json('/v1/party/groups', $filter);

		if (!isset($response['items'])) {
			yield from [];
		}

		foreach ($response['items'] as $item) {
			yield $this->convertGroup($item);
		}

		yield from [];
	}
}
