<?php

namespace OCA\GrueneGroupfolderManagement\Controller;

use OCA\GrueneGroupfolderManagement\Db\GroupMember;
use OCA\GrueneGroupfolderManagement\DTO\AddGroupMemberDto;
use OCA\GrueneGroupfolderManagement\DTO\CreateGroupDto;
use OCA\GrueneGroupfolderManagement\DTO\UpdateGroupDto;
use OCA\GrueneGroupfolderManagement\DTO\UpdateGroupMemberDto;
use OCA\GrueneGroupfolderManagement\Exceptions\NotFoundException;
use OCA\GrueneGroupfolderManagement\Manager\DivisionManager;
use OCA\GrueneGroupfolderManagement\Manager\FindGroupParams;
use OCA\GrueneGroupfolderManagement\Manager\GroupManager;
use OCA\GrueneGroupfolderManagement\Model\Division;
use OCP\AppFramework\Http\JSONResponse;

class GroupController extends BaseController {
	public function __construct(
		private GroupManager $groupManager,
		private DivisionManager $divisionManager,
	) {
		parent::__construct();
	}

	/**
	 * @NoAdminRequired
	 */
	public function create(array $createGroupDto) {
		/** @var CreateGroupDto */
		$createGroupDto = $this->validate($createGroupDto, CreateGroupDto::class);

		$division = $this->divisionManager->find($createGroupDto->divisionKey);
		if (!$division) {
			throw new NotFoundException(Division::class, $createGroupDto->divisionKey);
		}

		$this->denyAccessUnlessGranted(['CREATE_GROUP'], $division);

		$group = $this->groupManager->create($createGroupDto);
		$group = $this->groupManager->get($group->getId());

		return new JSONResponse($group);
	}

	/**
	 * @NoAdminRequired
	 */
	public function get(string $groupId) {
		$group = $this->groupManager->get($groupId);

		$this->denyAccessUnlessGranted(['READ_GROUP'], $group->getDivision());

		return new JSONResponse($group);
	}

	/**
	 * @NoAdminRequired
	 */
	public function find(int $offset, int $limit, string $divisions = '') {
		$params = new FindGroupParams(
			offset: max(0, $offset),
			limit: max(1, $limit),
			divisions: array_map('trim', explode(',', $divisions)),
		);

		$items = $this->groupManager->find($params);

		return new JSONResponse([
			'items' => $items,
		]);
	}

	/**
	 * @NoAdminRequired
	 */
	public function update(string $groupId, array $updateGroupDto) {
		/** @var UpdateGroupDto */
		$updateGroupDto = $this->validate($updateGroupDto, UpdateGroupDto::class);

		$group = $this->groupManager->get($groupId);

		$this->denyAccessUnlessGranted(['UPDATE_GROUP'], $group->getDivision());

		$group = $this->groupManager->update($group, $updateGroupDto);

		return new JSONResponse($group);
	}

	/**
	 * @NoAdminRequired
	 */
	public function delete(string $groupId) {
		$group = $this->groupManager->get($groupId);

		$this->denyAccessUnlessGranted(['DELETE_GROUP'], $group->getDivision());

		$group = $this->groupManager->delete($group);

		return new JSONResponse($group);
	}

	/**
	 * @NoAdminRequired
	 */
	public function findMemberChoices(string $search = '', ?int $groupId = null, int $limit = 20) {
		$choices = $this->groupManager->findMemberChoices(
			search: $search,
			groupId: $groupId,
			limit: $limit
		);

		return new JsonResponse([
			'items' => $choices,
		]);
	}

	/**
	 * @NoAdminRequired
	 */
	public function findMembers(string $groupId) {
		$members = $this->groupManager->findMembers($groupId);

		return new JsonResponse([
			'items' => array_map(fn($m) => $this->serializeGroupMember($m), $members)
		]);
	}

	/**
	 * @NoAdminRequired
	 */
	public function addMember(int $groupId, array $addGroupMemberDto) {
		$group = $this->groupManager->get($groupId);

		$this->denyAccessUnlessGranted(['UPDATE_GROUP'], $group->getDivision());
		$addGroupMemberDto = $this->validate($addGroupMemberDto, AddGroupMemberDto::class);

		$member = $this->groupManager->addMember($group, $addGroupMemberDto);

		return new JsonResponse($this->serializeGroupMember($member));
	}

	/**
	 * @NoAdminRequired
	 */
	public function updateMember(int $groupId, int $memberId, array $updateGroupMemberDto) {
		$group = $this->groupManager->get($groupId);

		$this->denyAccessUnlessGranted(['UPDATE_GROUP'], $group->getDivision());
		$updateGroupMemberDto = $this->validate($updateGroupMemberDto, UpdateGroupMemberDto::class);

		$member = $this->groupManager->updateMember($group, $memberId, $updateGroupMemberDto);

		return new JsonResponse($this->serializeGroupMember($member));
	}

	/**
	 * @NoAdminRequired
	 */
	public function deleteMember(int $groupId, int $memberId) {
		$group = $this->groupManager->get($groupId);

		$this->denyAccessUnlessGranted(['UPDATE_GROUP'], $group->getDivision());

		$member = $this->groupManager->deleteMember($group, $memberId);

		return new JsonResponse($member);
	}

	private function serializeGroupMember(GroupMember $member): array {
		$data = $member->jsonSerialize();

		$data["friendlyName"] = $member->getMappingId();

		if ('group' === $member->getMappingType()) {
			$sherpaMappingId = $this->groupManager->getSherpaName($member->getMappingId());

			if ($sherpaMappingId != $member->getMappingId()) {
				$data["mappingType"] = 'sherpa';
				$data["friendlyName"] = $sherpaMappingId;
			}
		}

		return $data;
	}
}
