<?php

declare(strict_types=1);

namespace OCA\GrueneGroupfolderManagement\DAV;

use OCA\DAV\Connector\Sabre\Node;
use OCA\Files_Trashbin\Sabre\AbstractTrash;
use OCA\GroupFolders\Mount\GroupMountPoint;
use OCA\GroupFolders\Trash\GroupTrashItem;
use OCA\GrueneGroupfolderManagement\Constants\Constants;
use OCA\GrueneGroupfolderManagement\Db\GroupMapper;
use OCA\GrueneGroupfolderManagement\Manager\GroupFolderManager;

use OCP\AppFramework\Db\DoesNotExistException;
use OCP\Files\Folder;
use OCP\Files\NotFoundException;

use Psr\Log\LoggerInterface;
use Sabre\DAV\Server;
use Sabre\DAV\ServerPlugin;

class GruenePlugin extends ServerPlugin {
	public const PROTECTED_PATH_LEVEL = 2;

	/** @var Server */
	protected $server;

	/** @var GroupMapper */
	private $groupMapper;

	/** @var GroupFolderManager */
	private $groupFolderManager;

	/** @var LoggerInterface */
	private $logger;

	public function __construct(
		GroupMapper $groupMapper,
		GroupFolderManager $groupFolderManager,
		LoggerInterface $logger
	) {
		$this->groupMapper = $groupMapper;
		$this->groupFolderManager = $groupFolderManager;
		$this->logger = $logger;
	}

	public function initialize(Server $server) {
		$this->server = $server;
		$this->server->on('beforeUnbind', [$this, 'beforeUnbind']);
		$this->server->on('beforeMove', [$this, 'beforeMove']);
		$this->server->on('beforeBind', [$this, 'beforeBind']);
	}


	/**
	 * Deny deleting and moving of gruene gliederungs folders
	 *
	 * @param string $path
	 * @return void
	 */
	public function beforeUnbind($path) {
		$node = $this->server->tree->getNodeForPath($path);
		if (!$node instanceof Node) {
			return;
		}
		$fileInfo = $node->getFileInfo();
		$mount = $fileInfo->getMountPoint();
		if (!$mount instanceof GroupMountPoint) {
			// ignore if the target file is not part of group folder storage
			return;
		}
		$nodePath = explode('/', trim($node->getPath(), '/'));
		if (count($nodePath) > self::PROTECTED_PATH_LEVEL) {
			// removing/moving nodes only allowed inside the gliederung
			return;
		}
		try {
			$this->groupMapper->getSingleDivisionByGroupFolder($nodePath[0], $nodePath[1]);
			// deny deleting/moving for all gliederungs folders
			return false;
		} catch (DoesNotExistException $e) {
			if (in_array($nodePath[1], Constants::DEFAULT_DIRECTORIES)) {
				// deny deleting/moving default directories
				return false;
			}
			return;
		}
	}


	/**
	 * Deny restoring nodes when gliederungs folder has been removed
	 * Deny moving nodes outside the group folder storage
	 *
	 * @param string $source
	 * @param string $destination
	 * @return void
	 */
	public function beforeMove($source, $destination) {
		$sourceNode = $this->server->tree->getNodeForPath($source);
		if (!$sourceNode instanceof AbstractTrash && !$sourceNode instanceof Node) {
			return;
		}
		$fileInfo = $sourceNode->getFileInfo();
		if ($fileInfo instanceof GroupTrashItem) {
			// in case of restoring file/folder into group folder storage
			$this->logger->info(
				sprintf(
					"Trying to restore file [%s] to original location [%s] from mount point [%s]",
					$fileInfo->getName(),
					$fileInfo->getOriginalLocation(),
					$fileInfo->getGroupFolderMountPoint()
				),
				['app' => 'gruenegroupfoldermanagement']
			);
			$original = $fileInfo->getOriginalLocation();
			$mount = $fileInfo->getGroupFolderMountPoint();
			$path = explode('/', $original);
			if (count($path) < self::PROTECTED_PATH_LEVEL) {
				return false;
			}
			if (in_array($path[0], Constants::DEFAULT_DIRECTORIES)) {
				return $this->restoreAncestorPath($original, $fileInfo->getTrashPath());
			}
			try {
				$this->groupMapper->getSingleDivisionByGroupFolder($mount, $path[0]);
				return $this->restoreAncestorPath($original, $fileInfo->getTrashPath());
			} catch (DoesNotExistException $e) {
				// deny restoring if gliederungs folder has been removed
				$this->logger->info(
					sprintf(
						"Denied restoring file [%s] original location [%s] mount point [%s], gruene folder [%s] does not exist",
						$fileInfo->getName(),
						$original,
						$mount,
						$path[0]
					),
					['app' => 'gruenegroupfoldermanagement']
				);
				return false;
			}
		} elseif ($fileInfo instanceof \OC\Files\FileInfo) {
			$mountPoint = $fileInfo->getMountPoint();
			if (!$mountPoint instanceof GroupMountPoint) {
				// ignore if the source file is not part of group folder storage
				return;
			}
			try {
				$destinationNode = $this->server->tree->getNodeForPath($destination);
			} catch (\Sabre\DAV\Exception\NotFound $e) {
				// Attempting to fetch its parent
				list($parentName, ) = \Sabre\Uri\split($destination);
				$destinationNode = $this->server->tree->getNodeForPath($parentName);
			}
			if (!$destinationNode instanceof Node) {
				return;
			}
			$destinationInfo = $destinationNode->getFileInfo();
			$destinationMount = $destinationInfo->getMountPoint();
			if (!$destinationMount instanceof GroupMountPoint) {
				// deny if destination is outside of group folder app
				return false;
			}
			$destinationPath = explode('/', trim($destinationNode->getPath(), '/'));
			if (count($destinationPath) < self::PROTECTED_PATH_LEVEL) {
				// deny if destination is outside the gliederungs folder
				return false;
			}
			return;
		}
		return;
	}


	/**
	 * Restores ancestores folders if they dont exist
	 *
	 * @param string $original
	 * @param string $trashPath
	 * @return void
	 */
	public function restoreAncestorPath(string $original, string $trashPath) {
		try {
			list($parent, $originalName) = \Sabre\Uri\split($original);
			$parentPath = explode('/', $parent);
			// extract group folder id from the trash path
			$groupFolderId = (int)explode('/', trim($trashPath, '/'))[0];
			$folder = $this->groupFolderManager->getGroupFolder($groupFolderId);
			while (true) {
				if (!count($parentPath) > 0) {
					try {
						# only allow restoring if there is no file with the same name at original location
						$folder->get($originalName);
						$this->logger->warning(
							"Could not restore file, original path: [$original] trash path: [$trashPath] group folder id: [$groupFolderId]
							file name already exists at original location",
							['app' => 'gruenegroupfoldermanagement']
						);
						return false;
					} catch (NotFoundException $e) {
						return;
					}
				}
				$directoryName = array_shift($parentPath);
				try {
					$folder = $folder->get($directoryName);
				} catch (NotFoundException $e) {
					/** @var Folder $folder */
					$folder = $folder->newFolder($directoryName);
					$this->logger->info(
						"Successfully restored ancestor path, original path: [$original] trash path: [$trashPath]
						group folder id: [$groupFolderId] restored directory: [$directoryName]",
						['app' => 'gruenegroupfoldermanagement']
					);
				}
			}
		} catch (\Throwable $e) {
			$this->logger->error(
				"Failed restoring ancestor path, original path: [$original] trash path: [$trashPath]. Error message: ".$e->getMessage(),
				['app' => 'gruenegroupfoldermanagement']
			);
			return false;
		}
	}


	/**
	 * Deny creating files in the group folders root space
	 * where only gliederungs folder may exist
	 *
	 * @param string $target
	 * @return void
	 */
	public function beforeBind($target) {
		list($parentName, ) = \Sabre\Uri\split($target);
		$node = $this->server->tree->getNodeForPath($parentName);
		if (!$node instanceof Node) {
			return;
		}
		$fileInfo = $node->getFileInfo();
		$mountPoint = $fileInfo->getMountPoint();
		if (!$mountPoint instanceof GroupMountPoint) {
			// ignore if the destination is not part of group folder storage
			return;
		}
		$nodePath = explode('/', trim($node->getPath(), '/'));
		if (!$fileInfo->getInternalPath()) {
			// deny if destination is outside the gliederungs folder
			return false;
		}
		return;
	}
}
