<?php

namespace OCA\GrueneGroupfolderManagement\Groups;

use OCA\GrueneGroupfolderManagement\Db\Group;
use OCA\GrueneGroupfolderManagement\Db\GroupMember;
use OCA\GrueneGroupfolderManagement\Db\GroupMapper;
use OCA\GrueneGroupfolderManagement\Db\GroupMemberMapper;
use OCA\GrueneGroupfolderManagement\Manager\SherpaManager;
use OCP\Group\Backend\ABackend;
use OCP\Group\Backend\ICountUsersBackend;
use OCP\GroupInterface;
use OCP\IGroupManager;
use OCP\IUserManager;

class GroupBackend extends ABackend implements ICountUsersBackend {
	/**
	 * @var IUserManager
	 */
	protected $user;

	/**
	 * @var IGroupManager
	 */
	protected $ociGroup;

	/**
	 * @var SherpaManager
	 */
	protected $sherpaManager;

	/**
	 * @var GroupMapper
	 */
	protected $groupMapper;

	/**
	 * @var GroupMemberMapper
	 */
	protected $groupMemberMapper;

	public function __construct(
		IUserManager $manager,
		IGroupManager $ociGroup,
		GroupMapper $groupMapper,
		GroupMemberMapper $groupMemberMapper
	) {
		$this->user = $manager;
		$this->ociGroup = $ociGroup;
		$this->groupMapper = $groupMapper;
		$this->groupMemberMapper = $groupMemberMapper;
	}


	protected function parseOrganizationGroupName(string $pattern, string $name, ?Group &$group) : bool {
		$matches = [];
		if (preg_match($pattern, $name, $matches)) {
			list($total, $organizationUuid, $groupId) = $matches;
			$found = $this->groupMapper->find($groupId);
			if ($found !== null && $found->getDivisionKey() == $organizationUuid) {
				$group = $found;
				return true;
			}
		}

		return false;
	}

	protected function parseOrganizationMemberGroupName(string $name, ?Group &$group) : bool {
		return $this->parseOrganizationGroupName(GroupName::MEMBER_GROUP_PATTERN, $name, $group);
	}

	protected function parseOrganizationAdminGroupName(string $name, ?Group &$group) : bool {
		return $this->parseOrganizationGroupName(GroupName::ADMIN_GROUP_PATTERN, $name, $group);
	}

	/**
	 * @param GroupMember[] $members
	 * @return string[]
	 */
	protected function covertGroupMembersToGids(?iterable $members) : array {
		if ($members === null) {
			return [];
		}

		$gids = [];

		foreach ($members as $member) {
			if ($member->getType() == GroupMemberMapper::TYPE_MEMBER) {
				$gids[] = GroupName::organizationMemberGroupName($member->getGroup());
			} elseif ($member->getType() == GroupMemberMapper::TYPE_ADMIN) {
				$gids[] = GroupName::organizationAdminGroupName($member->getGroup());
			}
		}

		return $gids;
	}

	protected function subgroups(Group $group, string $type, callable $mapper) {
		$groups = $this->covertGroupMembersToGids(
			$this->groupMemberMapper->findMembersInGroupByType($group, $type)
		);
		foreach ($groups as $subgroup) {
			foreach ($this->ociGroup->getBackends() as $backend) {
				$backendCheck = $backend;
				if ($backend instanceof CachedGroupBackendDecorator) {
					$backendCheck = $backend->getBackend();
				}
				if ($backendCheck instanceof self) {
					continue;
				}

				$return = $mapper($backend, $subgroup);
				if ($return !== null) {
					return $return;
				}
			}
		}

		return null;
	}


	/**
	 * is user in group?
	 * @param string $uid uid of the user
	 * @param string $gid gid of the group
	 * @return bool
	 * @since 4.5.0
	 *
	 * Checks whether the user is member of a group or not.
	 */
	public function inGroup($uid, $gid) {
		if ($gid === GroupName::PUBLIC_GROUP) {
			return true;
		}

		$check = function (GroupInterface $backend, $subgroup) use ($uid) {
			if ($backend->inGroup($uid, $subgroup)) {
				return true;
			}

			return null;
		};

		$group = null;
		if ($this->parseOrganizationMemberGroupName($gid, $group)) {
			if ($this->subgroups($group, GroupMemberMapper::TYPE_MEMBER, $check)) {
				return true;
			}

			$member = $this->groupMemberMapper->getGroupMember($group, 'user', $uid, GroupMemberMapper::TYPE_MEMBER);
			if ($member !== null) {
				return true;
			}
		}

		if ($this->parseOrganizationAdminGroupName($gid, $group)) {
			if ($this->subgroups($group, GroupMemberMapper::TYPE_ADMIN, $check)) {
				return true;
			}

			$member = $this->groupMemberMapper->getGroupMember($group, 'user', $uid, GroupMemberMapper::TYPE_ADMIN);
			if ($member !== null) {
				return true;
			}
		}

		return false;
	}


	/**
	 * Get all groups a user belongs to
	 * @param string $uid Name of the user
	 * @return array an array of group names
	 * @since 4.5.0
	 *
	 * This function fetches all groups a user belongs to. It does not check
	 * if the user exists at all.
	 */
	public function getUserGroups($uid) {
		if ($uid === null) {
			return [];
		}

		$implicitMemberships = [];
		foreach ($this->ociGroup->getBackends() as $backend) {
			$backendCheck = $backend;
			if ($backend instanceof CachedGroupBackendDecorator) {
				$backendCheck = $backend->getBackend();
			}
			if ($backendCheck instanceof self) {
				continue;
			}
			$implicitMemberships = array_merge($implicitMemberships, $backend->getUserGroups($uid));
		}
		$implicitMemberships = array_unique($implicitMemberships);

		$groups = [];
		$members = $this->groupMemberMapper->findGroupsByMappingIdAndType([$uid], 'user');
		if ($members !== null) {
			$groups = array_merge($groups, $this->covertGroupMembersToGids($members));
		}

		$members = $this->groupMemberMapper->findGroupsByMappingIdAndType($implicitMemberships, 'group');
		if ($groups !== null) {
			$groups = array_merge($groups, $this->covertGroupMembersToGids($members));
		}

		return array_unique(array_merge($groups, [
			GroupName::PUBLIC_GROUP
		]));
	}


	/**
	 * get a list of all groups
	 * @param string $search
	 * @param int $limit
	 * @param int $offset
	 * @return array an array of group names
	 * @since 4.5.0
	 *
	 * Returns a list with all groups
	 */
	public function getGroups($search = '', $limit = -1, $offset = 0) {
		$slimit = $limit;
		if ($limit == -1) {
			$slimit = null;
		}

		$soffset = $offset;
		if ($offset == 0) {
			$soffset = null;
		}

		$results = [];

		// TODO support partial search
		if (substr($search, 0, strlen(GroupName::GROUP_PREFIX)) === GroupName::GROUP_PREFIX) {
			/** @var Group $group */
			$group = null;
			if ($this->parseOrganizationMemberGroupName($search, $group)) {
				$results[] = GroupName::organizationMemberGroupName($group);
			} elseif ($this->parseOrganizationAdminGroupName($search, $group)) {
				$results[] = GroupName::organizationAdminGroupName($group);
			}
		} elseif ($search === '') {
			foreach ($this->groupMapper->findAll($slimit, $soffset) as $group) {
				$results[] = GroupName::organizationMemberGroupName($group);
				$results[] = GroupName::organizationAdminGroupName($group);
			}
		}

		if ($search === '' || strpos(GroupName::PUBLIC_GROUP, $search) !== false) {
			$results[] = GroupName::PUBLIC_GROUP;
		}

		return $results;
	}


	/**
	 * check if a group exists
	 * @param string $gid
	 * @return bool
	 * @since 4.5.0
	 */
	public function groupExists($gid) {
		$group = null;
		if ($this->parseOrganizationMemberGroupName($gid, $group)) {
			if ($group !== null) {
				return true;
			}
		}

		if ($this->parseOrganizationAdminGroupName($gid, $group)) {
			if ($group !== null) {
				return true;
			}
		}

		return $gid == GroupName::PUBLIC_GROUP;
	}

	/**
	 * get a list of all users in a group
	 * @param string $gid
	 * @param string $search
	 * @param int $limit
	 * @param int $offset
	 * @return array an array of user ids
	 * @since 4.5.0
	 */
	public function usersInGroup($gid, $search = '', $limit = -1, $offset = 0) {
		$pattern = "/^GROUP_[0-9]{8}_([0-9]+)_(?:MEMBERS|ADMINS)$/";
		preg_match($pattern, $gid, $matches);
		if (isset($matches[1])) {
			$groupsMembers = $this->groupMemberMapper->findMembersByGroup($matches[1]);
			$members = [];
			foreach ($groupsMembers as $member) {
				if ($member->getMappingType() === 'user') {
					$members[] = $member->getMappingId();
				} elseif ($member->getMappingType() === 'group') {
					// TODO add all users of this group to result
				}
			}
			return $members;
		}
		return [];
	}

	/**
	 * @since 14.0.0
	 */
	public function countUsersInGroup(string $gid, string $search = ''): int {
		// TODO repair

		if ($gid !== GroupName::PUBLIC_GROUP) {
			return 0;
		}

		if ($search === '') {
			$counts = $this->user->countUsers();
			return 0;
		}

		return 0;
	}
}
