<?php


namespace OCA\GrueneGroupfolderManagement\Groups;

use OCA\GrueneGroupfolderManagement\Constants\Constants;
use OCA\GrueneGroupfolderManagement\Db\Group;
use OCA\GrueneGroupfolderManagement\Model\DivisionGroup;
use OCA\GrueneGroupfolderManagement\Model\OrganizationalUnitInterface;
use OCA\GrueneGroupfolderManagement\Model\SAML\Division;
use OCA\GrueneGroupfolderManagement\Model\SAML\DivisionRole;
use OCA\GrueneGroupfolderManagement\Model\SAML\DivisionStaff;

class GroupName {
	public const PUBLIC_GROUP = 'public';

	public const DIVISION_PREFIX = 'GLIEDERUNG_';

	public const SHERPA_PREFIX = 'SHERPA_';

	public const GROUP_PREFIX = 'GROUP_';

	public const ADMIN_GROUP_SUFFIX = '_ADMINS';

	public const MEMBER_GROUP_SUFFIX = '_MEMBERS';

	public const TECHNICAL_WOLKE_SUFFIX = '_TECHNICAL_WOLKE';

	public const MANAGER_SUFFIX = '_MANAGER';

	public const EXECUTIVE_BOARD_SUFFIX = '_EXECUTIVE_BOARD';

	public const DIVISION_GROUP_PATTERN = '/^GLIEDERUNG_([0-9]{8})_MEMBERS$/';

	public const SHERPA_GROUP_PATTERN = '/^SHERPA_([0-9]{8})_([A-Z_]+)/';

	public const MEMBER_GROUP_PATTERN = '/^GROUP_([0-9]{8})_([0-9]+)_MEMBERS$/';

	public const ADMIN_GROUP_PATTERN = '/^GROUP_([0-9]{8})_([0-9]+)_ADMINS$/';

	public const DIVIVION_TYPE_LENGTH = 2;

	/**
	 * Returns the name of the division member group in the format "GLIEDERUNG_divisionKey_MEMBERS"
	 *
	 * @param OrganizationalUnitInterface $division
	 * @return string
	 */
	public static function divisionMember(OrganizationalUnitInterface $division) : string {
		return self::DIVISION_PREFIX . $division->getOrganizationalUuid() . self::MEMBER_GROUP_SUFFIX;
	}

	public static function saml(\OCA\GrueneGroupfolderManagement\Model\Division $division, $suffix = null) : string {
		$data = $division->getPrefix();

		if ($suffix !== null) {
			$data .= '_' . $suffix;
		}

		return $data;
	}

	public static function sherpa(DivisionGroup $division) : string {
		return self::SHERPA_PREFIX . $division->getDivision()->getOrganizationalUuid() . '_' . $division->getGroup()->getSlug();
	}

	public static function getTechnicalEditorGroupName(\OCA\GrueneGroupfolderManagement\Model\Division $division) {
		return self::SHERPA_PREFIX . $division->getOrganizationalUuid() . '_' . $division->getType() . self::TECHNICAL_WOLKE_SUFFIX;
	}

	public static function getManagerGroupName(\OCA\GrueneGroupfolderManagement\Model\Division $division) {
		return self::SHERPA_PREFIX . $division->getOrganizationalUuid() . '_' . $division->getType() . self::MANAGER_SUFFIX;
	}

	public static function getExecutiveBoardGroupName(\OCA\GrueneGroupfolderManagement\Model\Division $division) {
		return self::SHERPA_PREFIX . $division->getOrganizationalUuid() . '_' . $division->getType() . self::EXECUTIVE_BOARD_SUFFIX;
	}

	public static function parseSAMLRole(string $group) : iterable {
		$matches = [];
		if (preg_match('/^([0-9]{1,9})(_[0-9]+)?$/', $group, $matches)) {
			$gliederungsId = $matches[1];
			$rollenId = null;
			if (isset($matches[2])) {
				$rollenId = ltrim($matches[2], '_');
			}

			if ($rollenId !== null) {
				yield new DivisionRole($gliederungsId, $rollenId);
			} else {
				// in this case the user is staff member and not party member. We rewrite the division it that is prefixed
				// with 8 for staff members (therefore one digit longer) to a normal division number.
				// for some reasons they do not have a "Bundesverband" entry but just a "8" entry instead. In this case
				// we mock the "Bundesverband" entry
				if (preg_match('/^8/', $gliederungsId)) {
					$gliederungsId = substr($gliederungsId, 1);
					if ($gliederungsId === '') {
						$gliederungsId = '1';
					}
					yield new DivisionStaff($gliederungsId);
				} else {
					yield new Division($gliederungsId);
				}
			}
		}

		yield from [];
	}

	public static function organizationGroupPrefix(Group $group) : string {
		return self::GROUP_PREFIX . $group->getDivisionKey() . '_' . $group->getId();
	}

	public static function organizationGroupName(OrganizationalUnitInterface $division) : string {
		return $division->getOrganizationalUuid();
	}


	/**
	 * Returns the name of the group admins in the format "GROUP_divisionKey_groupId_ADMINS"
	 *
	 * @param Group $group
	 * @return string
	 */
	public static function organizationAdminGroupName(Group $group) {
		return self::organizationGroupPrefix($group) . self::ADMIN_GROUP_SUFFIX;
	}


	/**
	 * Returns the name of the group in the format "GROUP_divisionKey_groupId_MEMBERS"
	 *
	 * @param Group $group
	 * @return string
	 */
	public static function organizationMemberGroupName(Group $group) {
		return self::organizationGroupPrefix($group) . self::MEMBER_GROUP_SUFFIX;
	}

	/**
	 * get reserved directory names for this division
	 *
	 * @param \OCA\GrueneGroupfolderManagement\Model\Division $division
	 * @return string[] reserved directory names
	 */
	public static function getReservedNames(\OCA\GrueneGroupfolderManagement\Model\Division $division) {
		$divisionLevel = $division->getLevel();
		if ($divisionLevel === Constants::LOCAL_GROUP_LEVEL || $divisionLevel === Constants::COUNTY_CHAPTER_LEVEL) {
			return Constants::DEFAULT_DIRECTORIES;
		}
		return [];
	}

	public static function cleanName($name) {
		return str_replace('/', '_', $name);
	}
}
