<?php

namespace OCA\GrueneGroupfolderManagement\Groups;

use OCA\GrueneGroupfolderManagement\Manager\DivisionManager;
use OCA\GrueneGroupfolderManagement\Manager\SherpaManager;
use OCA\GrueneGroupfolderManagement\Model\Division;
use OCA\GrueneGroupfolderManagement\Model\DivisionGroup;
use OCP\Group\Backend\ABackend;
use OCP\IGroupManager;
use OCP\IUserManager;

class SherpaGroupBackend extends ABackend {
	/**
	 * @var IUserManager
	 */
	protected $userManager;

	/**
	 * @var SherpaManager
	 */
	protected $sherpaManager;

	/**
	 * @var DivisionManager
	 */
	protected $divisionManager;

	/**
	 * @var IGroupManager
	 */
	protected $groupManager;

	public function __construct(
		IGroupManager $groupManager,
		DivisionManager $divisionManager,
		SherpaManager $sherpaManager
	) {
		$this->groupManager = $groupManager;
		$this->divisionManager = $divisionManager;
		$this->sherpaManager = $sherpaManager;
	}

	public function parseSherpa(string $group, ?DivisionGroup &$divisionGroup) : bool {
		$matches = [];
		if (preg_match(GroupName::SHERPA_GROUP_PATTERN, $group, $matches)) {
			list($total, $organizationUuid, $sherpaSlug) = $matches;
			$division = $this->divisionManager->find($organizationUuid);
			if ($division === null) {
				return false;
			}

			$found = $this->sherpaManager->findByDivisionAndSlug($division, $sherpaSlug);

			if ($found !== null && $found->getDivision()->getOrganizationalUuid() == $organizationUuid) {
				$divisionGroup = $found;
				return true;
			}
		}

		return false;
	}

	public function parseDivision(string $group, ?Division &$division) : bool {
		$matches = [];
		if (preg_match(GroupName::DIVISION_GROUP_PATTERN, $group, $matches)) {
			list($total, $organizationUuid) = $matches;
			$division = $this->divisionManager->find($organizationUuid);
			if ($division !== null) {
				return true;
			}
		}

		return false;
	}

	/**
	 * is user in group?
	 * @param string $uid uid of the user
	 * @param string $gid gid of the group
	 * @return bool
	 * @since 4.5.0
	 *
	 * Checks whether the user is member of a group or not.
	 */
	public function inGroup($uid, $gid) {
		return in_array(
			$gid,
			$this->getUserGroups($uid)
		);
	}

	/**
	 * Get all groups a user belongs to
	 * @param string $uid Name of the user
	 * @return array an array of group names
	 * @since 4.5.0
	 *
	 * This function fetches all groups a user belongs to. It does not check
	 * if the user exists at all.
	 */
	public function getUserGroups($uid) {
		$gids = [];
		foreach ($this->groupManager->getBackends() as $backend) {
			$backendCheck = $backend;
			if ($backend instanceof CachedGroupBackendDecorator) {
				$backendCheck = $backend->getBackend();
			}

			if ($backendCheck instanceof self || $backendCheck instanceof GroupBackend) {
				continue;
			}

			$gids = array_merge($gids, $backend->getUserGroups($uid));
		}

		$groups = [];
		foreach ($this->sherpaManager->getDivisionRolesByGroupIds($gids) as $role) {
			$groups[] = GroupName::sherpa($role);
		}

		foreach ($this->sherpaManager->getDivisionsByGroupIds($gids) as $division) {
			$groups[] = GroupName::divisionMember($division);
		}

		return array_unique($groups);
	}

	/**
	 * get a list of all groups
	 * @param string $search
	 * @param int $limit
	 * @param int $offset
	 * @return array an array of group names
	 * @since 4.5.0
	 *
	 * Returns a list with all groups
	 */
	public function getGroups($search = '', $limit = -1, $offset = 0) {
		/** @var DivisionGroup $group */
		$group = null;
		if ($this->parseSherpa($search, $group)) {
			return [GroupName::sherpa($group)];
		}

		/** @var Division $division */
		$division = null;
		if ($this->parseDivision($search, $division)) {
			return [GroupName::divisionMember($division)];
		}

		return [];
	}

	/**
	 * check if a group exists
	 * @param string $gid
	 * @return bool
	 * @since 4.5.0
	 */
	public function groupExists($gid) {
		/** @var Division $division */
		/** @var DivisionGroup $group */
		$group = $division = null;
		if ($this->parseSherpa($gid, $group)) {
			return true;
		}

		if ($this->parseDivision($gid, $division)) {
			return true;
		}

		return false;
	}

	/**
	 * get a list of all users in a group
	 * @param string $gid
	 * @param string $search
	 * @param int $limit
	 * @param int $offset
	 * @return array an array of user ids
	 * @since 4.5.0
	 */
	public function usersInGroup($gid, $search = '', $limit = -1, $offset = 0) {
		$users = [];

		/** @var DivisionGroup $group */
		$group = null;
		if ($this->parseSherpa($gid, $group)) {
			$samlGroups = [];
			foreach ($group->getGroup()->getSherpaIds() as $sherpaId) {
				$samlGroups[] = GroupName::saml($group->getDivision(), $sherpaId);
			}

			foreach ($samlGroups as $samlGroup) {
				foreach ($this->groupManager->getBackends() as $backend) {
					$backendCheck = $backend;
					if ($backend instanceof CachedGroupBackendDecorator) {
						$backendCheck = $backend->getBackend();
					}

					if ($backendCheck instanceof self || $backendCheck instanceof GroupBackend) {
						continue;
					}

					$users = array_merge($users, $backend->usersInGroup($samlGroup, $search, $limit, $offset));
				}
			}
		}

		/** @var Division $division */
		$division = null;
		if ($this->parseDivision($gid, $division)) {
			/*
			if ($division->getId() == '11002601') {
				$users[] = 'RutaMaxi';
			}
			*/

			// TODO search all members
		}

		return array_unique($users);
	}
}
