<?php

namespace OCA\GrueneGroupfolderManagement\Listener;

use OCA\DAV\Events\CalendarShareUpdatedEvent;
use OCA\Files_Sharing\Exceptions\SharingRightsException;
use OCA\GrueneGroupfolderManagement\Constants\Constants;
use OCA\GrueneGroupfolderManagement\Groups\GroupName;
use OCP\IDBConnection;
use OCP\IGroupManager;
use OCP\IUserManager;
use OCP\EventDispatcher\Event;
use OCP\EventDispatcher\IEventListener;
use OCP\Share\Events\ShareCreatedEvent;
use OCP\Share\IManager;
use OCP\Share\IShare;

use Sabre\DAV\Exception\Forbidden;

class ShareListener implements IEventListener {
	/** @var IManager */
	private $shareManager;

	/** @var IUserManager */
	private $userManager;

	/** @var IGroupManager */
	private $groupManager;

	/** @var IDBConnection */
	private $connection;

	public function __construct(IManager $shareManager, IUserManager $userManager, IGroupManager $groupManager, IDBConnection $connection) {
		$this->shareManager = $shareManager;
		$this->userManager = $userManager;
		$this->groupManager = $groupManager;
		$this->connection = $connection;
	}

	/**
	 * @param Event $event
	 * @return void
	 */
	public function handle(Event $event): void {
		if ($event instanceof ShareCreatedEvent) {
			$this->handleFileShareUpdate($event);
		}
		if ($event instanceof CalendarShareUpdatedEvent) {
			$this->handleCalendarShareUpdate($event);
		}
	}

	/**
	 * @param ShareCreatedEvent $event
	 */
	private function handleFileShareUpdate(ShareCreatedEvent $event) {
		$share = $event->getShare();
		if ($share->getShareType() !== IShare::TYPE_GROUP) {
			return;
		}
		if (!preg_match('/^([0-9]{1,8})$/', $share->getSharedWith(), $matches) || !isset($matches[1])) {
			// ignore if shared with other groups than gliederungs groups
			return;
		}
		$this->checkSharePrivileg($share->getSharedBy(), $matches[1], $share);
	}

	/**
	 * Parses user, who shares this calendar and the target to share with
	 *
	 * @param CalendarShareUpdatedEvent $event
	 */
	private function handleCalendarShareUpdate(CalendarShareUpdatedEvent $event) {
		$sharedWithGroups = [];
		foreach ($event->getAdded() as $added) {
			if (!isset($added["href"])) {
				continue;
			}
			// check if share target is a group referencing a complete division
			if (!preg_match('/^principal:principals\/groups\/([0-9]{1,8})$/', $added["href"], $matches) || !isset($matches[1])) {
				continue;
			}
			$sharedWithGroups[] = $matches[1];
		}
		if (empty($sharedWithGroups)) {
			return;
		}
		// if there is more than one target group to share
		if (count($sharedWithGroups) > 1) {
			// sort division ids by key length ascending to check just highest ranked division
			usort($sharedWithGroups, function (string $divisionA, string $divisionB) {
				return strlen($divisionA) - strlen($divisionB);
			});
		}
		$calendarData = $event->getCalendarData();
		// parse user, who shares this calendar
		if (
			!isset($calendarData["principaluri"])
			|| !preg_match('/^principals\/users\/([\w]+)$/', $calendarData["principaluri"], $matches)
			|| !isset($matches[1])
		) {
			return;
		}
		$this->checkSharePrivileg($matches[1], $sharedWithGroups[0], null, $calendarData["id"]);
	}

	/**
	 * Only technical wolke members can share to all members of a division
	 * BV: can share to all other divisions
	 * LV: can share to LV, KV, OV
	 * KV: can share to KV, OV
	 * OV: can share to OV
	 *
	 * @param string $sharedBy
	 * @param string $sharedWithDivisionId
	 * @param IShare|null $share
	 * @return void
	 */
	private function checkSharePrivileg(string $sharedBy, string $sharedWithDivisionId, IShare $share = null, int $calendarId = null) {
		$sharedBy = $this->userManager->get($sharedBy);
		$userGroupIds = $this->groupManager->getUserGroupIds($sharedBy);
		// as userGroupIds contains already the pseudo sherpa groups, resolving groups from api is not required
		$technicalWolkeGroups = preg_grep('/^SHERPA_[0-9]{1,8}_(BV|LV|KV|OV)_TECHNICAL_WOLKE$/', $userGroupIds);
		if (!$technicalWolkeGroups) {
			// only technical wolke role is allowed to share with whole division group
			if (!is_null($share)) {
				$this->shareManager->deleteShare($share);
				throw new SharingRightsException("Teilen mit allen Nutzern einer Gliederung ist nicht gestattet");
			} elseif (!is_null($calendarId)) {
				$this->deleteCalendarShares($calendarId, $sharedWithDivisionId);
				throw new Forbidden("Teilen mit allen Nutzern einer Gliederung ist nicht gestattet");
			}
		}
		// OV is the minimum technical wolke group
		$maxDivisionRank = Constants::OV_KEY_LENGTH;
		// a user may be member of more than one technical wolke group
		foreach ($technicalWolkeGroups as $technicalWolkeGroup) {
			// extract division type from technical wolke group
			$divisionType = substr(
				$technicalWolkeGroup,
				strlen($technicalWolkeGroup) - (strlen(GroupName::TECHNICAL_WOLKE_SUFFIX) + GroupName::DIVIVION_TYPE_LENGTH),
				GroupName::DIVIVION_TYPE_LENGTH
			);
			$divisionIdLength = $this->getDivisionIdLength($divisionType);
			// only the maximum technical wolke group is important
			$maxDivisionRank = $divisionIdLength < $maxDivisionRank ? $divisionIdLength: $maxDivisionRank;
		}
		if ($maxDivisionRank > strlen($sharedWithDivisionId)) {
			# if the shared with division has higher rank than users technical wolke rank, sharing is forbidden
			if (!is_null($share)) {
				$this->shareManager->deleteShare($share);
				throw new SharingRightsException("Teilen mit allen Nutzern einer übergeordneten Gliederung ist nicht gestattet");
			} elseif (!is_null($calendarId)) {
				$this->deleteCalendarShares($calendarId, $sharedWithDivisionId);
				throw new Forbidden("Teilen mit allen Nutzern einer übergeordneten Gliederung ist nicht gestattet");
			}
		}
	}

	/**
	 * @param string $divisionType
	 * @return int
	 */
	private function getDivisionIdLength(string $divisionType) {
		switch ($divisionType) {
			case Constants::BV_TYPE:
				return Constants::BV_KEY_LENGTH;
			case Constants::LV_TYPE:
				return Constants::LV_KEY_LENGTH;
			case Constants::KV_TYPE:
				return Constants::KV_KEY_LENGTH;
			default:
				return Constants::OV_KEY_LENGTH;
		}
	}

	/**
	 * @param array $calendarData
	 */
	private function deleteCalendarShares(int $calendarId, string $divisionId) {
		$delete = $this->connection->getQueryBuilder();
		$delete->delete('dav_shares')
			->where($delete->expr()->eq('principaluri', $delete->createNamedParameter("principals/groups/$divisionId")))
			->andWhere($delete->expr()->eq('resourceid', $delete->createNamedParameter($calendarId)));
		$delete->execute();
	}
}
