<?php

namespace OCA\GrueneGroupfolderManagement\Manager;

use OCA\GrueneGroupfolderManagement\Groups\GroupName;
use OCA\GrueneGroupfolderManagement\Constants\Constants as GrueneConst;
use OCA\GrueneGroupfolderManagement\Db\Group;
use OCA\GrueneGroupfolderManagement\Model\Division;
use OCP\Constants;
use Psr\Log\LoggerInterface;
use OCP\Files\Folder;
use OCP\Files\NotFoundException;

class DirectoryManager {
	/** @var GroupFolderManager */
	protected $groupFolderManager;

	/** @var MappingManager */
	protected $mappingManager;

	/** @var LoggerInterface */
	protected $logger;

	public function __construct(GroupFolderManager $groupFolderManager, MappingManager $mappingManager, LoggerInterface $logger) {
		$this->groupFolderManager = $groupFolderManager;
		$this->mappingManager = $mappingManager;
		$this->logger = $logger;
	}


	/**
	 * Get or create group folder and save applicable groups and ACL permissions
	 *
	 * @param Group[] $group
	 * @return void
	 */
	public function saveToFileSystem(iterable $groups, Division $division, $ensureRootPermissions = false) {
		$mountPoint = GroupName::cleanName($division->getOrganizationalName());
		$divisionGroupName = GroupName::divisionMember($division);
		$rootGroupFolderColumns = $exists = $this->groupFolderManager->getGroupFolderByOrganizationKey($division->getId());
		if (empty($rootGroupFolderColumns)) {
			$rootGroupFolderId = $this->groupFolderManager->createGroupFolderByMountPoint($mountPoint, $divisionGroupName);
			$this->logger->info(
				sprintf('Could not find gruene groupfolder [%s] and created new one', $mountPoint),
				['app' => 'gruenegroupfoldermanagement']
			);
		} else {
			$rootGroupFolderId = (int)$rootGroupFolderColumns['folder_id'];
			if ($rootGroupFolderColumns['mount_point'] !== $mountPoint) {
				$this->groupFolderManager->updateMountPoint($rootGroupFolderId, $mountPoint);
			}
		}
		$rootGroupFolder = $this->groupFolderManager->getGroupFolder($rootGroupFolderId);
		if (!$exists || $ensureRootPermissions) {
			$this->grantGroupFolderPermissions($rootGroupFolder, $division, $rootGroupFolderId);
			if (in_array($division->getLevel(), [GrueneConst::COUNTY_CHAPTER_LEVEL, GrueneConst::LOCAL_GROUP_LEVEL])) {
				// TODO consider proper api fetching for default directory group names like SHERPA_10100700_KV_MANAGER
				// $sherpaGroups = $this->sherpaManager->getDivisionRolesByDivisionAndTypes($division, GrueneConst::DEFAULT_GROUPS);
				$this->ensureDefaultDirectories(
					$rootGroupFolder,
					GrueneConst::EXECUTIVE_BOARD,
					$rootGroupFolderId,
					[GroupName::getExecutiveBoardGroupName($division)]
				);
				$this->ensureDefaultDirectories(
					$rootGroupFolder,
					GrueneConst::EXECUTIVE_BOARD_MANAGER,
					$rootGroupFolderId,
					[
						GroupName::getExecutiveBoardGroupName($division),
						GroupName::getManagerGroupName($division)
					]
				);
			}
		}
		foreach ($groups as $group) {
			$this->ensureFolderAclRules($group, $rootGroupFolder, $rootGroupFolderId);
		}
	}


	/**
	 * Grant division members access to root group folder
	 * Enable ACL rules for entire root group folder
	 * Set ACL rules on root group folder level
	 *
	 * @param Folder $rootGroupFolder
	 * @param Division $division
	 * @param integer $rootGroupFolderId
	 * @return void
	 */
	private function grantGroupFolderPermissions(Folder $rootGroupFolder, Division $division, int $rootGroupFolderId) {
		$folderId = $rootGroupFolder->getId();
		$technicalEditor = GroupName::getTechnicalEditorGroupName($division);
		$divisionGroupName = GroupName::divisionMember($division);
		$this->groupFolderManager->grantGroupFolderPermissionToGroup($rootGroupFolderId, GroupName::divisionMember($division));
		$this->groupFolderManager->enableAclRulesForGroupFolder($rootGroupFolderId, true);
		// restrict division members to read access at root space
		$this->groupFolderManager->saveAcl(
			$divisionGroupName,
			$folderId,
			GroupFolderManager::NO_INHERITANCE_MASK,
			Constants::PERMISSION_READ
		);
		// let technical editor clear group folder root space
		$this->groupFolderManager->saveAcl(
			$technicalEditor,
			$folderId,
			GroupFolderManager::NO_INHERITANCE_MASK,
			Constants::PERMISSION_UPDATE | Constants::PERMISSION_DELETE
		);
	}


	/**
	 * Get or create division folder and save ACL rules
	 *
	 * @param Group $group
	 * @param Folder $rootGroupFolder
	 * @param integer $rootGroupFolderId
	 * @return void
	 */
	private function ensureFolderAclRules(Group $group, Folder $rootGroupFolder, int $rootGroupFolderId) {
		$subFolderId = $this->getOrCreateDivisionFolder($group, $rootGroupFolder, $rootGroupFolderId);
		if ($group->getActive()) {
			$divisionAclPermission = $group->getDivisionAclPermission();
			$memberAclPermission = $group->getMemberAclPermission();
			$adminAclPermission = $group->getAdminAclPermission();
			$divisionAclMask = $group->getDivisionAclMask();
			$memberAclMask = $group->getMemberAclMask();
			$adminAclMask = $group->getAdminAclMask();
		} else {
			// disable this division group = revoke all permission
			$divisionAclPermission = $memberAclPermission = $adminAclPermission = GroupFolderManager::DENY_ALL_PERMISSION;
			// never inherit from group folder permissions
			$divisionAclMask = $memberAclMask = $adminAclMask = GroupFolderManager::NO_INHERITANCE_MASK;
		}
		$aclRules = [
			[
				'groupName' => GroupName::divisionMember($group->getDivision()),
				'aclPermission' => $divisionAclPermission,
				'aclMask' => $divisionAclMask
			],
			[
				'groupName' => GroupName::organizationMemberGroupName($group),
				'aclPermission' => $memberAclPermission,
				'aclMask' => $memberAclMask
			],
			[
				'groupName' => GroupName::organizationAdminGroupName($group),
				'aclPermission' => $adminAclPermission,
				'aclMask' => $adminAclMask
			],
			[
				'groupName' => 'public',
				'aclPermission' => GroupFolderManager::DENY_ALL_PERMISSION,
				'aclMask' => GroupFolderManager::NO_INHERITANCE_MASK
			],
		];
		$this->setAclRulesForDivisionGroup($subFolderId, $aclRules, $rootGroupFolderId, $rootGroupFolder->getId());
	}


	/**
	 * Setup a default directory inside the group folder and grant access to sherpa groups
	 *
	 * @param Folder $rootGroupFolder
	 * @param string $directoryName
	 * @param integer $rootGroupFolderId
	 * @param array $sherpaGroups
	 * @return void
	 */
	private function ensureDefaultDirectories(Folder $rootGroupFolder, string $directoryName, int $rootGroupFolderId, array $sherpaGroups) {
		try {
			$folder = $rootGroupFolder->get($directoryName);
		} catch (NotFoundException $e) {
			$folder = $rootGroupFolder->newFolder($directoryName);
			$this->logger->info(
				sprintf('Could not find any existing gruene default folder and created a new one [%s]', $directoryName),
				['app' => 'gruenegroupfoldermanagement']
			);
		}
		$aclRules = [
			[
				'groupName' => 'public',
				'aclPermission' => GroupFolderManager::DENY_ALL_PERMISSION,
				'aclMask' => GroupFolderManager::NO_INHERITANCE_MASK
			]
		];
		foreach ($sherpaGroups as $sherpaGroup) {
			$aclRules[] = [
				'groupName' => $sherpaGroup,
				'aclPermission' => GroupFolderManager::ALLOW_ALL_PERMISSION,
				'aclMask' => GroupFolderManager::NO_INHERITANCE_MASK
			];
		}
		$this->setAclRulesForDivisionGroup($folder->getId(), $aclRules, $rootGroupFolderId, $rootGroupFolder->getId());
	}


	/**
	 * Find child node by id
	 *
	 * @param Folder $folder
	 * @param string $id
	 * @return \OCP\Files\Node|null
	 */
	protected function findById(Folder $folder, string $id) {
		$folders = $folder->getDirectoryListing();
		foreach ($folders as $item) {
			if ($item->getId() == $id) {
				return $item;
			}
		}
		return null;
	}


	/**
	 * Get the folder id beneath the root group folder or create the folder
	 *
	 * @param Group $directory
	 * @param Folder $rootGroupFolder
	 *
	 * @return int
	 */
	private function getOrCreateDivisionFolder(Group $divisionGroup, Folder $rootGroupFolder, int $rootGroupFolderId) {
		$divisionId = $divisionGroup->getDivision()->getOrganizationalUuid();
		// fetch folder id from database
		$newFolderId = $previousFolderId = (int)$this->mappingManager->fetchForeignKey($divisionId, 'group_' . $divisionGroup->getId());
		$directoryName = GroupName::cleanName($divisionGroup->getName());
		$folder = null;
		if ($newFolderId !== null) {
			// try to find folder by id
			$folder = $this->findById($rootGroupFolder, $newFolderId);
		}
		if ($folder === null) {
			// if the folder could not be found by id
			try {
				// try to find folder by group name
				$folder = $rootGroupFolder->get($directoryName);
				$this->logger->info(
					sprintf(
						'Could not find any existing gruene folder id [%s] for division [%s], but found one by name',
						$directoryName,
						$divisionGroup->getDivisionKey()
					),
					['app' => 'gruenegroupfoldermanagement']
				);
			} catch (NotFoundException $e) {
				// create a new folder for this  group name
				$folder = $rootGroupFolder->newFolder($directoryName);
				$this->logger->info(
					sprintf(
						'Could not find any existing gruene folder and created a new one [%s] for division [%s]',
						$directoryName,
						$divisionGroup->getDivisionKey()
					),
					['app' => 'gruenegroupfoldermanagement']
				);
			}
		}
		$newFolderId = $folder->getId();
		$previousFolderName = $folder->getName();
		if ($previousFolderName !== $directoryName) {
			// if editor renamed this group, update folder name
			$storage = $this->groupFolderManager->getStorage($rootGroupFolder, $rootGroupFolderId);
			$storage->rename($previousFolderName, $directoryName);
			$storage->getUpdater()->renameFromStorage($storage, $previousFolderName, $directoryName);
			$this->logger->info(
				sprintf('Renaming gruene folder [%s] to [%s] was successful for division [%s]',
					$previousFolderName,
					$directoryName,
					$divisionGroup->getDivisionKey()
				),
				['app' => 'gruenegroupfoldermanagement']
			);
		}
		if ($previousFolderId !== $newFolderId) {
			// update the folder ids database entry
			$this->mappingManager->setForeignKey($newFolderId, $divisionId, 'group_' . $divisionGroup->getId());
			$this->logger->info(
				sprintf(
					'Updated gruenegfm_mapping folder id successfully for this gruene folder [%s] of division [%s]',
					$directoryName,
					$divisionGroup->getDivisionKey()
				),
				['app' => 'gruenegroupfoldermanagement']
			);
		}
		return $newFolderId;
	}


	/**
	 * Set ACL rules for single folder and grant access to root group folder
	 *
	 * @param int $folderId target folder to grant access
	 * @param iterable $rules
	 * @param int $rootGroupFolderId id of a group folder
	 * @param int $rootFolderId id of the Folder instance
	 */
	private function setAclRulesForDivisionGroup(int $folderId, array $rules, int $rootGroupFolderId, int $rootFolderId) {
		$groupNames = [];
		foreach ($rules as $rule) {
			$this->groupFolderManager->saveAcl(
				$rule['groupName'],
				$folderId,
				$rule['aclMask'],
				$rule['aclPermission']
			);
			if ($rule['groupName'] !== 'public') {
				// restrict groups folder access to read-only on root level
				$this->groupFolderManager->saveAcl(
					$rule['groupName'],
					$rootFolderId,
					GroupFolderManager::NO_INHERITANCE_MASK,
					Constants::PERMISSION_READ
				);
				$groupNames[] = $rule['groupName'];
			}
		}
		// grant groups access to root group folder
		foreach ($groupNames as $groupName) {
			$this->groupFolderManager->grantGroupFolderPermissionToGroup(
				$rootGroupFolderId,
				$groupName
			);
		}
	}


	/**
	 * Deletes the folder and its related ACL rules
	 *
	 * @param Group $group
	 * @return void
	 */
	public function deleteFolder(Group $group) {
		// get group folder db entry
		$groupFolderColumns = $this->groupFolderManager->getGroupFolderByOrganizationKey($group->getDivision()->getId());
		$groupFolderId = (int)$groupFolderColumns['folder_id'];
		// get groupfolder directory, groupfolder id needed
		$groupFolder = $this->groupFolderManager->getGroupFolder($groupFolderId, false);
		$this->groupFolderManager->clearTrashBin(
			$groupFolderId, GroupName::cleanName($group->getName())
		);
		// the folder to be removed by name
		$folder = $groupFolder->get(GroupName::cleanName($group->getName()));
		// remove all applicable group folders groups for this group folder
		$groupIds = array(
			GroupName::organizationMemberGroupName($group),
			GroupName::organizationAdminGroupName($group)
		);
		$this->groupFolderManager->removeApplicableGroup($groupFolderId, $groupIds);
		$this->groupFolderManager->deleteAcl($folder->getId(), $groupIds);
		// delete entry in mapping database
		$this->mappingManager->deleteFolder($folder->getId());
		$folder->delete();
	}
}
