<?php

namespace OCA\GrueneGroupfolderManagement\Manager;

use OC\Files\Storage\Wrapper\Jail;
use OCA\GroupFolders\ACL\Rule;
use OCA\GroupFolders\ACL\RuleManager;
use OCA\GroupFolders\ACL\UserMapping\UserMappingManager;
use OCA\GroupFolders\Folder\FolderManager;
use OCA\GroupFolders\Helper\LazyFolder;
use OCA\GroupFolders\Mount\MountProvider;
use OCA\GroupFolders\Mount\GroupFolderStorage;
use OCP\IDBConnection;
use OCP\DB\QueryBuilder\IQueryBuilder;
use OCP\Files\Folder;
use OCP\Files\NotFoundException;
use OCP\IUserSession;
use Psr\Log\LoggerInterface;
use Throwable;

class GroupFolderManager {
	public const FALLBACK_QUOTA = 5368709120;
	public const DENY_ALL_PERMISSION = 0;
	public const ALLOW_ALL_PERMISSION = 31;
	public const NO_INHERITANCE_MASK = 31;

	/** @var FolderManager */
	protected $folderManager;

	/** @var IDBConnection */
	protected $db;

	/** @var MountProvider */
	protected $mountProvider;

	/** @var RuleManager */
	protected $ruleManager;

	/** @var UserMappingManager */
	protected $userMappingManager;

	/** @var LazyFolder */
	protected $appFolder;

	/** @var LoggerInterface */
	protected $logger;

	/** @var IUserSession */
	protected $userSession;

	public function __construct(
		MountProvider $mountProvider,
		FolderManager $folderManager,
		RuleManager $ruleManager,
		UserMappingManager $userMappingManager,
		IDBConnection $connection,
		LazyFolder $appFolder,
		LoggerInterface $logger,
		IUserSession $userSession
	) {
		$this->folderManager = $folderManager;
		$this->db = $connection;
		$this->mountProvider = $mountProvider;
		$this->ruleManager = $ruleManager;
		$this->userMappingManager = $userMappingManager;
		$this->appFolder = $appFolder;
		$this->logger = $logger;
		$this->userSession = $userSession;
	}


	/**
	 * Get database entry for this mount point
	 *
	 * @param string $divisionMountPoint
	 * @return array|false
	 */
	public function getGroupFolderByMountPoint(string $mountPoint) {
		$query = $this->db->getQueryBuilder();
		$query->select('*')
			->from('gruenegfm_group_folders')
			->where($query->expr()->eq('mount_point', $query->createNamedParameter($mountPoint)));
		$statement = $query->executeQuery();
		$groupFolder = $statement->fetch();
		return $groupFolder;
	}

	public function getGroupFolderByOrganizationKey(string $organizationKey) {
		$query = $this->db->getQueryBuilder();
		$query->select('*')
			->from('gruenegfm_group_folders')
			->where('organization_key = :organizationKey')
			->setParameter('organizationKey', $organizationKey);
		$statement = $query->executeQuery();
		$groupFolder = $statement->fetch();
		return $groupFolder;
	}


	/**
	 * Get database entry for this folder id
	 *
	 * @param int $folderId
	 * @return array|false
	 */
	public function getGroupFolderByFolderId(int $folderId) {
		$query = $this->db->getQueryBuilder();
		$query->select('*')
			->from('gruenegfm_group_folders')
			->where($query->expr()->eq('folder_id', $query->createNamedParameter($folderId)));
		$statement = $query->execute();
		$groupFolder = $statement->fetch();
		return $groupFolder;
	}


	/**
	 * Create new database entry for this mount point
	 *
	 * @param string $mountPoint
	 * @return void
	 */
	public function createGroupFolderByMountPoint(string $mountPoint) {
		$folderId = $this->folderManager->createFolder($mountPoint);
		$this->folderManager->setFolderQuota(
			$folderId,
			self::FALLBACK_QUOTA
		);
		return $folderId;
	}

	public function updateMountPoint(int $folderId, string $mountPoint) {
		try {
			$this->folderManager->setMountPoint($folderId, $mountPoint);
		} catch (Throwable $e) {
			$this->logger->error("Failed to update mount point for folder ID ${$folderId}: " . $e->getMessage());
		}
	}


	/**
	 * @param int $folderId
	 * @param bool $enabled
	 */
	public function enableAclRulesForGroupFolder(int $folderId, bool $enabled) {
		$this->folderManager->setFolderACL($folderId, $enabled);
	}


	/**
	 * Grant permissions for the whole group folder
	 *
	 * @param int $folderId
	 * @param string $groupId
	 * @param int    $permission
	 */
	public function grantGroupFolderPermissionToGroup(int $folderId, string $groupId, int $permission = 31) {
		try {
			$this->folderManager->addApplicableGroup($folderId, $groupId);
		} catch (\Exception $e) {
			// already added
		}
		$this->folderManager->setGroupPermissions($folderId, $groupId, $permission);
	}


	/**
	 * Removes all applicable groups for a single group folder
	 *
	 * @param int $folderId
	 * @param iterable $groupIds
	 * @return void
	 */
	public function removeApplicableGroup(int $folderId, iterable $groupIds) {
		$query = $this->db->getQueryBuilder();
		$query->delete('group_folders_groups')
			->where($query->expr()->eq('folder_id', $query->createNamedParameter($folderId, IQueryBuilder::PARAM_INT)))
			->andWhere($query->expr()->in('group_id', $query->createNamedParameter($groupIds, IQueryBuilder::PARAM_STR_ARRAY)));
		$query->execute();
	}


	/**
	 * @param int $folderId
	 * @param bool $create
	 *
	 * @return Folder
	 */
	public function getGroupFolder(int $folderId, bool $create = true) {
		// get or create directory for this group folder database entry
		$folder = $this->mountProvider->getFolder($folderId, $create);
		if ($folder === null) {
			throw new \RuntimeException(sprintf('Could not create folder with id: [%d]', $folderId));
		}
		return $folder;
	}


	/**
	 * @param Folder
	 * @return GroupFolderStorage
	 */
	public function getStorage(Folder $rootGroupFolder, int $folderId) {
		$cacheEntry = $this->appFolder->getStorage()->getCache()->get($rootGroupFolder->getId());
		$rootPath = $this->mountProvider->getJailPath($folderId);
		$baseStorage = new Jail([
			'storage' => $rootGroupFolder->getStorage(),
			'root' => $rootPath
		]);
		return new GroupFolderStorage([
			'storage' => $baseStorage,
			'folder_id' => $folderId,
			'mountOwner' => null,
			'userSession' => $this->userSession,
			'rootCacheEntry' => $cacheEntry,
			'quota' => null
		]);
	}


	/**
	 * Removes all ACL rules of deleted folder
	 * Also removes ACL rules of default groups like GROUP_10300000_groupid_MEMBERS
	 *
	 * @param integer $fileId
	 * @param array $mappingIds
	 * @return void
	 */
	public function deleteAcl(int $fileId, array $mappingIds) {
		$query = $this->db->getQueryBuilder();
		$query->delete('group_folders_acl')
			->where($query->expr()->in('mapping_id', $query->createNamedParameter($mappingIds, IQueryBuilder::PARAM_STR_ARRAY)))
			->orWhere($query->expr()->eq('fileid', $query->createNamedParameter($fileId)));
		$query->execute();
	}


	/**
	 * Saves ACL rule if valid
	 *
	 * @param string $mappingId
	 * @param integer $fileId
	 * @param integer $mask
	 * @param integer $permission
	 * @return void
	 */
	public function saveAcl(string $mappingId, int $fileId, int $mask, int $permission) {
		$mapping = $this->userMappingManager->mappingFromId('group', $mappingId);
		if ($mapping === null) {
			$this->logger->error(
				sprintf('Could not find group mapping for id [%s] and hence could not create ACL rule', $mappingId),
				['app' => 'gruenegroupfoldermanagement']
			);
			return null;
		}
		$rule = new Rule($mapping, $fileId, $mask, $permission);
		if ($rule === null) {
			$this->logger->error(
				sprintf('Could not initialize ACL rule for group [%s]', $mappingId),
				['app' => 'gruenegroupfoldermanagement']
			);
			return;
		}
		$this->ruleManager->saveRule($rule);
	}


	/**
	 * Clear trash bin for this gruene folder
	 *
	 * @param int $groupFolderId
	 * @param string $grueneFolderName
	 * @return void
	 */
	public function clearTrashBin(int $groupFolderId, string $grueneFolderName) {
		$this->logger->info(
			"Start clearing trash for group folder [$groupFolderId] and gruene folder [$grueneFolderName]",
			['app' => 'gruenegroupfoldermanagement']
		);
		$trashRoot = $this->getTrashFolder($groupFolderId);
		$trashItems = $this->getTrashItems($groupFolderId, $grueneFolderName);
		$allTrashItems = $trashItems ? implode(",", array_map(function ($trashItem) {
			return $trashItem['name'];
		}, $trashItems)) : "None";
		$this->logger->info("Clearing trash items [$allTrashItems] for group folder [$groupFolderId] and gruene folder [$grueneFolderName]",
			['app' => 'gruenegroupfoldermanagement']
		);
		foreach ($trashItems as $trashItem) {
			$trashNode = $trashRoot->get($trashItem['name'] . '.d' . $trashItem['deleted_time']);
			if ($trashNode === null) {
				$this->logger->error(
					sprintf('Failed to remove item from trashbin [%s]', $trashItem['name']),
					['app' => 'gruenegroupfoldermanagement']
				);
				continue;
			}
			if ($trashNode->getStorage()->unlink($trashNode->getInternalPath()) === false) {
				$this->logger->error(
					sprintf('Failed to remove item from trashbin [%s]', $trashItem['name']),
					['app' => 'gruenegroupfoldermanagement']
				);
				continue;
			}
			$trashNode->getStorage()->getCache()->remove($trashNode->getInternalPath());
		}
		$this->removeTrashItems($groupFolderId, $grueneFolderName);
		$this->logger->info(
			"Successfully cleared trash for group folder [$groupFolderId] and gruene folder [$grueneFolderName]",
			['app' => 'gruenegroupfoldermanagement']
		);
	}


	/**
	 * Get groupfolders root trash directory
	 *
	 * @param integer $folderId
	 * @return Folder
	 */
	private function getTrashFolder(int $folderId) {
		try {
			return $this->appFolder->get('trash/' . $folderId);
		} catch (NotFoundException $e) {
			/** @var Folder $trashRoot */
			$trashRoot = $this->appFolder->nodeExists('trash') ? $this->appFolder->get('trash') : $this->appFolder->newFolder('trash');
			return $trashRoot->newFolder((string)$folderId);
		}
	}


	/**
	 * Remove database trash entries for this gruene folder
	 *
	 * @param integer $groupFolderId
	 * @param string $grueneFolderName
	 * @return void
	 */
	private function removeTrashItems(int $groupFolderId, string $grueneFolderName) {
		$query = $this->db->getQueryBuilder();
		$query->delete('group_folders_trash')
			->where($query->expr()->eq('folder_id', $query->createNamedParameter($groupFolderId, IQueryBuilder::PARAM_INT)))
			->andWhere(
				$query->expr()->like(
					'original_location',
					$query->createNamedParameter($this->db->escapeLikeParameter($grueneFolderName) . '/%')
				)
			);
		$query->execute();
	}


	/**
	 * Get all trash entries for this gruene folder
	 *
	 * @param integer $groupFolderId
	 * @param string $grueneFolderName
	 * @return array
	 */
	private function getTrashItems(int $groupFolderId, string $grueneFolderName) {
		$query = $this->db->getQueryBuilder();
		$query->select('*')
			->from('group_folders_trash')
			->where($query->expr()->eq('folder_id', $query->createNamedParameter($groupFolderId, IQueryBuilder::PARAM_INT)))
			->andWhere(
				$query->expr()->like(
					'original_location',
					$query->createNamedParameter($this->db->escapeLikeParameter($grueneFolderName) . '/%')
				)
			);
		$statement = $query->execute();
		$result = $statement->fetchAll();
		return $result;
	}
}
