<?php

namespace OCA\GrueneGroupfolderManagement\Manager;

use OCA\GrueneGroupfolderManagement\Constants\Constants;
use OCA\GrueneGroupfolderManagement\Db\Group;
use OCA\GrueneGroupfolderManagement\Db\GroupMapper;
use OCA\GrueneGroupfolderManagement\Db\GroupMember;
use OCA\GrueneGroupfolderManagement\Db\GroupMemberMapper;
use OCA\GrueneGroupfolderManagement\DTO\AddGroupMemberDto;
use OCA\GrueneGroupfolderManagement\DTO\CreateGroupDto;
use OCA\GrueneGroupfolderManagement\DTO\UpdateGroupDto;
use OCA\GrueneGroupfolderManagement\DTO\UpdateGroupMemberDto;
use OCA\GrueneGroupfolderManagement\Exceptions\ConflictedStateException;
use OCA\GrueneGroupfolderManagement\Exceptions\NotFoundException;
use OCA\GrueneGroupfolderManagement\Groups\SherpaGroupBackend;
use OCA\GrueneGroupfolderManagement\Model\Division;
use OCA\GrueneGroupfolderManagement\Model\DivisionGroup;
use OCA\GrueneGroupfolderManagement\Exceptions\ValidationFailedException;
use OCP\IGroup;
use OCP\IGroupManager;
use OCP\IUser;
use OCP\IUserManager;
use Psr\Log\LoggerInterface;

class GroupManager {
	public function __construct(
		private IUserManager $ocpUserManager,
		private IGroupManager $ocpGroupManager,
		private GroupMapper $groupMapper,
		private GroupMemberMapper $groupMemberMapper,
		private DivisionManager $divisionManager,
		private DirectoryManager $directoryManager,
		private SherpaGroupBackend $sherpaGroupBackend,
		private LoggerInterface $logger,
	) {
	}

	public function create(CreateGroupDto $createGroupDto): Group {
		$division = $this->divisionManager->find($createGroupDto->divisionKey);

		if (!$division) {
			throw new NotFoundException('Division', ['id' => $createGroupDto->divisionKey]);
		}

		$group = $this->groupMapper->findByDivisionAndName($createGroupDto->divisionKey, $createGroupDto->name);

		if ($group) {
			throw new ConflictedStateException('group with this name already exists');
		}

		$group = new Group();

		// default values need to be present for DirectoryManager::ensureFolderAclRules
		$group->setActive(true);
		$group->setDivisionAclMask(\OCP\Constants::PERMISSION_ALL);
		$group->setDivisionAclPermission(1);
		$group->setMemberAclMask(\OCP\Constants::PERMISSION_ALL);
		$group->setMemberAclPermission(1);
		$group->setAdminAclMask(\OCP\Constants::PERMISSION_ALL);
		$group->setAdminAclPermission(1);

		$group->setDivision($division);

		$group->setFields($createGroupDto);

		$this->validateGroupName($group);

		$group = $this->groupMapper->insert($group);

		$this->directoryManager->saveToFileSystem([$group], $division);

		return $group;
	}

	/**
	 * @throws NotFoundException
	 */
	public function get(int $groupId): Group {
		$group = $this->groupMapper->find($groupId);

		if (!$group) {
			throw new NotFoundException(Group::class, ['id' => $groupId]);
		}

		return $group;
	}

	public function find(FindGroupParams $params) {
		return $this->groupMapper->findMany(
			offset: $params->offset,
			limit: $params->limit,
			divisions: $params->divisions,
		);
	}

	public function update(Group $group, UpdateGroupDto $updateGroupDto): Group {
		$group->setFields($updateGroupDto);
		$this->validateGroupName($group);

		$group = $this->groupMapper->update($group);

		$this->directoryManager->saveToFileSystem([$group], $group->getDivision());

		return $group;
	}

	public function delete(Group $group): Group {
        try {
            $this->directoryManager->deleteFolder($group);
            $this->groupMemberMapper->deleteMembersByGroupId($group->getId());
            $this->groupMapper->delete($group);
        } catch (\Throwable $e) {
            $this->logger->error(
                sprintf(
                    'Failed deleting gruene folder [%s] of division [%s] with error message [%s]',
                    $group->getName(),
                    $group->getDivisionKey(),
                    $e->getMessage()
                ),
                ['app' => 'gruenegroupfoldermanagement']
            );
			throw new \RuntimeException('delete of group failed');
        }

		return $group;
	}

	/**
	 * @return GroupMember[]
	 */
	public function findMembers(int $groupId): array {
		$members = $this->groupMemberMapper->findMany(
			groupId: $groupId,
			limit: 200,
		);

		return $members;
	}

	/**
	 * Look for member choices for a group to add
	 * @return GroupMember[]
	 */
	public function findMemberChoices(
		string $search,
		?int $groupId,
		?int $limit = 20,
		): array {
		$users = $this->ocpUserManager->search($search, $limit);
		$groups = $this->ocpGroupManager->search($search, $limit);

		$mapUser = fn(IUser $user) => [
			'type' => 'user',
			'id' => $user->getUID(),
			'displayName' => $user->getDisplayName(),
			'email' => $user->getEMailAddress(),
		];

		$mapGroup = fn(IGroup $group) => [
			'type' => 'group',
			'id' => $group->getGID(),
			'displayName' => $group->getDisplayName(),
		];

		$choices = [
			...array_map($mapUser, array_values($users)),
			...array_map($mapGroup, array_values($groups)),
		];

		// filter out groups and users which are already member of group
		if ($groupId) {
			$members = $this->findMembers($groupId);
			$choices = array_filter($choices, function($choice) use ($members) {
				foreach ($members as $member) {
					if ($choice['type'] === $member->getMappingType() && $choice['id'] === $member->getMappingId()) {
						return false;
					}
				}
				return true;
			});
		};

		return $choices;
	}

	public function addMember(Group $group, AddGroupMemberDto $dto): GroupMember {
		$member = $this->groupMemberMapper->findOne(
			groupId: $group->getId(),
			mappingId: $dto->mappingId,
			mappingType: $dto->mappingType,
		);

		if ($member) {
			throw new ConflictedStateException('member already exists');
		}

		$member = new GroupMember();
		$member->setGroupId($group->getId());
		$member->setMappingId($dto->mappingId);
		$member->setMappingType($dto->mappingType);
		$member->setType($dto->type);

		$member = $this->groupMemberMapper->insert($member);

		return $member;
	}

	public function updateMember(
		Group $group,
		int $memberId,
		UpdateGroupMemberDto $dto
		): GroupMember {
		$member = $this->groupMemberMapper->get($memberId);

		if (!$member) {
			throw new NotFoundException(GroupMember::class, ['id' => $memberId]);
		}

		if ($dto->type) {
			$member->setType($dto->type);
		}

		$member = $this->groupMemberMapper->update($member);

		return $member;
	}


	public function deleteMember(
		Group $group,
		int $memberId,
		): GroupMember {
		$member = $this->groupMemberMapper->get($memberId);

		if (!$member) {
			throw new NotFoundException(GroupMember::class, ['id' => $memberId]);
		}

		$member = $this->groupMemberMapper->delete($member);

		return $member;
	}

	public function getSherpaName(string $name) : string {
		$object = null;
		if ($this->sherpaGroupBackend->parseSherpa($name, $object) && $object instanceof DivisionGroup) {
			return $object->getDivision()->getOrganization()->getName() . ' / ' . $object->getDivision()->getType() . ' ' . $object->getDivision()->getName() /* . ' (' . $object->getDivision()->getPrefix() . ')' */ . ' / ' . $object->getGroup()->getName() /* . ' ('.$object->getGroup()->getType().')' */;
		} elseif ($this->sherpaGroupBackend->parseDivision($name, $object) && $object instanceof Division) {
			return $object->getOrganization()->getName() . ' / ' . $object->getType() . ' ' . $object->getName() /*. ' (' . $object->getPrefix() . ')' */ . ' / Mitglied';
		}

		return $name;
	}

	private function validateGroupName(Group $group) {
		if (
			in_array($group->getDivision()->getLevel(), [
				Constants::LOCAL_GROUP_LEVEL,
				Constants::COUNTY_CHAPTER_LEVEL
			]) &&
			in_array(mb_strtolower($group->getName()), Constants::DEFAULT_DIRECTORIES)
		) {
			throw new ValidationFailedException('Invalid group name', [
				'name' => [
					'message' => 'name contains reserved keyword',
				],
			]);
		}
	}
}
